//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

// MACROS ------------------------------------------------------------------

#define GRAVITY			1225.0
#define MAXMOVE			1050.0
#define STOPSPEED		2.1875
#define FRICTION		3.28125
#define FRICTION_WATER	3.0

#define BOUNCE_VAL		1.5

// CONSTANTS ---------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// FUNCTION PROTOTYPES -----------------------------------------------------

// DATA DECLARATIONS -------------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	ClipVelocity
//
//	Slide off of the impacting object
//
//==========================================================================

TVec ClipVelocity(TVec in, TVec normal, float overbounce)
{
	return in - normal * (DotProduct(in, normal) * overbounce);
}

//**************************************************************************
//
//	SLIDE MOVE
//
//	Allows the player to slide along any angled walls.
//
//**************************************************************************

float			bestslidefrac;
line_t*			bestslideline;

mobj_t*			slidemo;
TVec			slideorg;
TVec			slidedir;

//==========================================================================
//
//	PTR_SlideTraverse
//
//==========================================================================

boolean PTR_SlideTraverse(intercept_t* in)
{
	line_t*		li;
	TVec		hit_point;
	opening_t	*open;

    if (!in->isaline)
		Error("PTR_SlideTraverse: not a line?");
		
    li = in->line;

    if (li->flags & ML_TWOSIDED)
    {
	    // set openrange, opentop, openbottom
		hit_point = slideorg + in->frac * slidedir;
    	open = LineOpenings(li, hit_point);
		open = FindOpening(open, slidemo->origin.z,	slidemo->origin.z + slidemo->height);
    
	    if (open &&
	    	(open->range >= slidemo->height) &&	//	fits
    		(open->top - slidemo->origin.z >= slidemo->height) && // mobj is not too high
	    	(open->bottom - slidemo->origin.z <= 24.0)) // not too big a step up
		{
    		// this line doesn't block movement
    		return true;
		}
	}
	else
    {
		if (PointOnPlaneSide(slidemo->origin, li))
		{
	    	// don't hit the back side
	    	return true;
		}
    }

    // the line does block movement,
    // see if it is closer than best so far
    if (in->frac < bestslidefrac)
    {
		bestslidefrac = in->frac;
		bestslideline = li;
    }
	
    return false;	// stop
}

//==========================================================================
//
//	SlidePathTraverse
//
//==========================================================================

void SlidePathTraverse(float x, float y)
{
	slideorg = vector(x, y, slidemo->origin.z);
	slidedir = slidemo->velocity * frametime;
	P_PathTraverse(x, y, x + slidedir.x, y + slidedir.y,
		PT_ADDLINES, PTR_SlideTraverse);
}

//==========================================================================
//
//	P_SlideMove
//
//	The momx / momy move is bad, so try to slide along a wall.
//	Find the first line hit, move flush to it, and slide along it.
//	This is a kludgy mess.
//
//==========================================================================

void P_SlideMove(mobj_t* mo)
{
	float		leadx;
	float		leady;
	float		trailx;
	float		traily;
	float		newx;
	float		newy;
    int			hitcount;
		
    slidemo = mo;
    hitcount = 0;

	do
    {
	    if (++hitcount == 3)
		{
			// don't loop forever
			if (!P_TryMove(mo, mo->origin.x, mo->origin.y + mo->velocity.y * frametime))
			    P_TryMove(mo, mo->origin.x + mo->velocity.x * frametime, mo->origin.y);
			return;
		}

	    // trace along the three leading corners
	    if (mo->velocity.x > 0.0)
    	{
			leadx = mo->origin.x + mo->radius;
			trailx = mo->origin.x - mo->radius;
    	}
	    else
    	{
			leadx = mo->origin.x - mo->radius;
			trailx = mo->origin.x + mo->radius;
	    }

	    if (mo->velocity.y > 0.0)
	    {
			leady = mo->origin.y + mo->radius;
			traily = mo->origin.y - mo->radius;
	    }
    	else
	    {
			leady = mo->origin.y - mo->radius;
			traily = mo->origin.y + mo->radius;
	    }
		
	    bestslidefrac = 1.00001;

	    SlidePathTraverse(leadx, leady);
	    SlidePathTraverse(trailx, leady);
	    SlidePathTraverse(leadx, traily);
    
	    // move up to the wall
    	if (bestslidefrac == 1.00001)
	    {
			// the move most have hit the middle, so stairstep
			if (!P_TryMove(mo, mo->origin.x, mo->origin.y + mo->velocity.y * frametime))
			    P_TryMove(mo, mo->origin.x + mo->velocity.x * frametime, mo->origin.y);
			return;
    	}

	    // fudge a bit to make sure it doesn't hit
    	bestslidefrac -= 0.03125;
	    if (bestslidefrac > 0.0)
    	{
			newx = mo->velocity.x * frametime * bestslidefrac;
			newy = mo->velocity.y * frametime * bestslidefrac;
	
			if (!P_TryMove(mo, mo->origin.x + newx, mo->origin.y + newy))
            {
				if (!P_TryMove(mo, mo->origin.x, mo->origin.y + mo->velocity.y * frametime))
				    P_TryMove(mo, mo->origin.x + mo->velocity.x * frametime, mo->origin.y);
				return;
			}
	    }
    
	    // Now continue along the wall.
    	// First calculate remainder.
	    bestslidefrac = 1.0 - (bestslidefrac + 0.03125);
    
	    if (bestslidefrac > 1.0)
			bestslidefrac = 1.0;
    
	    if (bestslidefrac <= 0.0)
			return;
    
	    // clip the moves
		mo->velocity = ClipVelocity(mo->velocity * bestslidefrac, bestslideline->normal, 1.0);

    } while (!P_TryMove(mo, mo->origin.x + mo->velocity.x * frametime,
    	mo->origin.y + mo->velocity.y * frametime));
}

//**************************************************************************
//
//	OBJECT MOVEMENT
//
//**************************************************************************

//==========================================================================
//
//	CheckWater
//
//==========================================================================

boolean CheckWater(mobj_t *mobj)
{
	TVec	point;
	int		cont;

	point = mobj->origin;
	point.z += 1.0;
	
	mobj->waterlevel = 0;
	mobj->watertype = CONTENTS_EMPTY;
	cont = PointContents(mobj->subsector->sector, point);
	if (cont >= CONTENTS_WATER)
	{
		mobj->watertype = cont;
		mobj->waterlevel = 1;
		point.z = mobj->origin.z + mobj->height * 0.5;
		cont = PointContents(mobj->subsector->sector, point);
		if (cont >= CONTENTS_WATER)
		{
			mobj->waterlevel = 2;
			if (mobj->player)
			{
				point = mobj->player->vieworg;
				cont = PointContents(mobj->subsector->sector, point);
				if (cont >= CONTENTS_WATER)
					mobj->waterlevel = 3;
			}
		}
	}
	return mobj->waterlevel > 1;
}

//==========================================================================
//
//	P_XYMovement
//
//==========================================================================

void P_XYMovement(mobj_t* mo)
{ 	
    float		ptryx;
    float		ptryy;
    float		xmove;
    float		ymove;
			
    if (!mo->velocity.x && !mo->velocity.y)
    {
		if (mo->flags & MF_SKULLFLY)
		{
	    	// the skull slammed into something
		    mo->flags &= ~MF_SKULLFLY;
		    mo->velocity.x = mo->velocity.y = mo->velocity.z = 0.0;

		    SetMobjState(mo, mo->spawnstate);
		}
		return;
    }
	
    if (mo->velocity.x > MAXMOVE)
		mo->velocity.x = MAXMOVE;
    else if (mo->velocity.x < -MAXMOVE)
		mo->velocity.x = -MAXMOVE;

    if (mo->velocity.y > MAXMOVE)
		mo->velocity.y = MAXMOVE;
    else if (mo->velocity.y < -MAXMOVE)
		mo->velocity.y = -MAXMOVE;
		
    xmove = mo->velocity.x * frametime;
    ymove = mo->velocity.y * frametime;

    do
    {
		if (xmove > MAXMOVE / (35.0 * 2.0) || ymove > MAXMOVE / (35.0 * 2.0))
		{
		    ptryx = mo->origin.x + xmove / 2.0;
		    ptryy = mo->origin.y + ymove / 2.0;
		    xmove /= 2.0;
		    ymove /= 2.0;
		}
		else
		{
		    ptryx = mo->origin.x + xmove;
	    	ptryy = mo->origin.y + ymove;
		    xmove = ymove = 0.0;
		}
		
		if (!P_TryMove(mo, ptryx, ptryy))
		{
	    	// blocked move
		    if (mo->player)
		    {
		    	// try to slide along it
				P_SlideMove(mo);
		    }
		    else if (mo->flags & MF_MISSILE)
		    {
				// explode a missile
				if (ceilingline &&
			    	ceilingline->backsector &&
				    ceilingline->backsector->ceiling.pic == skyflatnum)
				{
				    // Hack to prevent missiles exploding
			    	// against the sky.
				    // Does not handle sky floors.
				    P_RemoveMobj(mo);
				    return;
				}
				P_ExplodeMissile(mo);
		    }
		    else
			{
				mo->velocity.x = mo->velocity.y = 0.0;
			}
		}
    } while (xmove || ymove);
}

//==========================================================================
//
// P_ZMovement
//
//==========================================================================

void P_ZMovement(mobj_t* mo)
{
	float	dist;
    float	delta;

    // check for smooth step up
    if (mo->player && mo->origin.z < mo->floorz)
    {
		mo->player->viewheight -= mo->floorz - mo->origin.z;

		mo->player->deltaviewheight
		    = (VIEWHEIGHT - mo->player->viewheight) / 8.0;
    }
    
    // adjust height
    mo->origin.z += mo->velocity.z * frametime;
	
    if (mo->flags & MF_FLOAT && mo->target)
    {
		// float down towards target if too close
		if (!(mo->flags & MF_SKULLFLY) && !(mo->flags & MF_INFLOAT))
		{
		    dist = MobjDist2(mo, mo->target);
	    
		    delta = mo->target->origin.z + mo->height / 2.0 - mo->origin.z;

		    if (delta < 0.0 && dist < -(delta * 3.0))
				mo->origin.z -= FLOATSPEED * frametime;
		    else if (delta > 0.0 && dist < (delta * 3.0))
				mo->origin.z += FLOATSPEED * frametime;
		}
    }
    
    // clip movement
    if (mo->origin.z <= mo->floorz)
    {
		// hit the floor

		// Note (id):
		//  somebody left this after the setting momz to 0,
		//  kinda useless there.
		if (mo->flags & MF_SKULLFLY)
		{
		    // the skull slammed into something
		    mo->velocity.z = -mo->velocity.z;
		}
	
		if (mo->velocity.z < 0.0)
		{
		    if (mo->player && (mo->velocity.z < -GRAVITY * 0.25))
		    {
				// Squat down.
				// Decrease viewheight for a moment
				// after hitting the ground (hard),
				// and utter appropriate sound.
				mo->player->deltaviewheight = mo->velocity.z / 8.0 * frametime;
				StartSoundName(mo, "PlayerLand");
		    }
		}
		mo->origin.z = mo->floorz;

		if (mo->flags & MF_MISSILE)
		{
			if (!(mo->flags & MF_NOCLIP))
			{
			    P_ExplodeMissile(mo);
		    	return;
			}
		}
    }

    if (mo->origin.z + mo->height > mo->ceilingz)
    {
		// hit the ceiling
		if (mo->velocity.z > 0.0)
	    	mo->velocity.z = 0.0;

	    mo->origin.z = mo->ceilingz - mo->height;

		if (mo->flags & MF_SKULLFLY)
		{
			// the skull slammed into something
		    mo->velocity.z = -mo->velocity.z;
		}

		if (mo->flags & MF_MISSILE)
		{
			if (!(mo->flags & MF_NOCLIP))
			{
			    P_ExplodeMissile(mo);
	    		return;
			}
		}
    }
}

//==========================================================================
//
//	UpdateVelocity
//
//==========================================================================

void UpdateVelocity(mobj_t *mo)
{
	float		dot;
    player_t*	player;

    player = mo->player;

	if (mo->waterlevel < 2)
	{
	    if (!(mo->flags & MF_NOGRAVITY) &&
    		(mo->origin.z > mo->floorz || mo->subsector->sector->floor.normal.z < 0.7))
	    {
			//	Add gravity
	    	mo->velocity.z -= GRAVITY * frametime;
	    }
	}
	else if (!mo->player)
	{
	    if (!(mo->flags & MF_NOGRAVITY) &&
    		(mo->origin.z > mo->floorz || mo->subsector->sector->floor.normal.z < 0.7))
	    {
			//	Add gravity
	    	mo->velocity.z = -60.0;
	    }
	}
	
    // slow down

    if (mo->flags & (MF_MISSILE | MF_SKULLFLY))
		return; 	// no friction for missiles ever
		
    if (mo->origin.z > mo->floorz && mo->waterlevel < 2)
		return;		// no friction when airborne

	//	Clip velocity
	if (mo->origin.z <= mo->floorz)
	{
		dot = DotProduct(mo->velocity, mo->floor->normal);
		if (dot < 0.0)
		{
			mo->velocity -= dot * mo->floor->normal;
		}
	}

    if (mo->flags & MF_CORPSE)
    {
		// do not stop sliding if halfway off a step with some momentum
		if (mo->velocity.x > 0.25 * 35.0 || mo->velocity.x < -0.25 * 35.0 ||
			mo->velocity.y > 0.25 * 35.0 || mo->velocity.y < -0.25 * 35.0)
		{
		    if (mo->floorz != mo->subsector->sector->floorheight)
				return;
		}
    }

    if (mo->velocity.x > -STOPSPEED && mo->velocity.x < STOPSPEED &&
		mo->velocity.y > -STOPSPEED && mo->velocity.y < STOPSPEED &&
		mo->velocity.z > -STOPSPEED && mo->velocity.z < STOPSPEED &&
    	(!player || (!player->forwardmove && !player->sidemove)))
    {
		if (player)
		{
        	int		sn;

			// if in a walking frame, stop moving
            sn = player->mo->statenum;
			if (sn >= S_PLAY_RUN1 && sn <= S_PLAY_RUN4)
		    	SetMobjState(player->mo, S_PLAY);
		}
		mo->velocity = vector(0.0, 0.0, 0.0);
    }
    else
    {
		if (mo->waterlevel > 1)
			mo->velocity -= mo->velocity * (FRICTION_WATER * frametime);
		else
			mo->velocity -= mo->velocity * (FRICTION * frametime);
    }
}

//==========================================================================
//
//	P_MobjThinker
//
//==========================================================================

void P_MobjThinker(mobj_t* mobj)
{
 	//  In my deathmatch mode remove dropped weapons after 30 seconds
    // because otherwise they are spawned fastar than player(s) can pick 
    // them up and the game is going slower and slower and slower ...
	if ((mobj->flags & MF_DROPPED) && (deathmatch == 3))
	{
       	mobj->movecount++;
        if (mobj->movecount >= 30 * 35)
		{
           	P_RemoveMobj(mobj);
            return;
		}
	}

	CheckWater(mobj);
	UpdateVelocity(mobj);

    // momentum movement
    if (mobj->velocity.x || mobj->velocity.y || (mobj->flags & MF_SKULLFLY))
    {
		P_XYMovement(mobj);

		if (!mobj->real_function)
        {
	    	return;		// mobj was removed
		}
    }
    if ((mobj->origin.z != mobj->floorz) || mobj->velocity.z)
    {
		P_ZMovement(mobj);

		if (!mobj->real_function)
        {
	    	return;		// mobj was removed
		}
    }

    // cycle through states,
    // calling action functions at transitions
    if (mobj->time != -1.0)
    {
		mobj->time -= frametime;
		
		// you can cycle through multiple states in a tic
		if (mobj->time <= 0.0)
		{
			mobj->time = 0.0;
	    	if (!SetMobjState(mobj, mobj->nextstate))
				return;		// freed itself
		}
    }
    else
    {
		// check for nightmare respawn
		if (!(mobj->flags & MF_COUNTKILL))
	    	return;

        //  In my deathmatch mode monstars respawns very fast
        if (deathmatch == 3)
		{
			if (level->tictime & 31)
	    		return;

			if (P_Random() > 32)
	    		return;

			P_NightmareRespawn(mobj);
		}
        else
		{
			if (!respawnmonsters)
	    		return;

        	mobj->movecount++;

			if (mobj->movecount < 12 * 35)
	    		return;

			if (level->tictime & 31)
	    		return;

			if (P_Random() > 4)
	    		return;

			P_NightmareRespawn(mobj);
		}
    }
}

