//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	BSP traversal, handling of LineSegs for rendering.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "r_local.h"

// MACROS ------------------------------------------------------------------

//#define TEST_LIGHTS
//#define NEED_AXIS

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

TVec				r_saxis;
TVec				r_taxis;
TVec				r_texorg;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static sector_t*	frontsector;
static sector_t*	backsector;

static int			r_frustum_indexes[4][6];

static sec_region_t	*r_region;
static sec_plane_t	*r_floor;
static sec_plane_t	*r_ceiling;
static sec_plane_t	*r_back_floor;
static sec_plane_t	*r_back_ceiling;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	CalcLineLights
//
//==========================================================================

#ifdef TEST_LIGHTS
void CalcLineLights(line_t *line, float vl0, float vl1, float vl2, float vl3)
{
	float linelen = Length(*line->v2 - *line->v1);
	for (int i = 0; i < cl_level.numsegs; i++)
	{
		seg_t *seg = &cl_level.segs[i];
		if (seg->linedef != line)
		{
			continue;
		}
		if (seg->side == 0)
		{
			seg->vlight[0] = vl0 + (vl3 - vl0) * seg->offset / linelen;
			seg->vlight[1] = vl1 + (vl2 - vl1) * seg->offset / linelen;
			seg->vlight[2] = vl1 + (vl2 - vl1) * (seg->offset + seg->length) / linelen;
			seg->vlight[3] = vl0 + (vl3 - vl0) * (seg->offset + seg->length) / linelen;
		}
		else
		{
			seg->vlight[0] = vl3 + (vl0 - vl3) * seg->offset / linelen;
			seg->vlight[1] = vl2 + (vl1 - vl2) * seg->offset / linelen;
			seg->vlight[2] = vl2 + (vl1 - vl2) * (seg->offset + seg->length) / linelen;
			seg->vlight[3] = vl3 + (vl0 - vl3) * (seg->offset + seg->length) / linelen;
		}
	}
}
#endif

//==========================================================================
//
//	R_SetUpFrustumIndexes
//
//==========================================================================

static void R_SetUpFrustumIndexes(void)
{
	for (int i = 0; i < 4; i++)
	{
		int *pindex = r_frustum_indexes[i];
		for (int j = 0; j < 3; j++)
		{
			if (view_clipplanes[i].normal[j] < 0)
			{
				pindex[j] = j;
				pindex[j + 3] = j + 3;
			}
			else
			{
				pindex[j] = j + 3;
				pindex[j + 3] = j;
			}
		}
	}
}

//==========================================================================
//
//	UpdateNodeBBox
//
//==========================================================================

static void UpdateNodeBBox(int bspnum, float *bbox)
{
    if (bspnum & NF_SUBSECTOR)
    {
		int num;

		if (bspnum == -1)
		    num = 0;
		else
		    num = bspnum & (~NF_SUBSECTOR);
	    sector_t *sec = cl_level.subsectors[num].sector;
		bbox[2] = sec->floor.minz;
		if (sec->ceiling.pic == skyflatnum)
			bbox[5] = sec->skyheight;
		else
			bbox[5] = sec->ceiling.maxz;
	}
	else
	{
		node_t* bsp = &cl_level.nodes[bspnum];
    
    	UpdateNodeBBox(bsp->children[0], bsp->bbox[0]);
		UpdateNodeBBox(bsp->children[1], bsp->bbox[1]);
		bbox[2] = MIN(bsp->bbox[0][2], bsp->bbox[1][2]);
		bbox[5] = MAX(bsp->bbox[0][5], bsp->bbox[1][5]);
	}
}

//==========================================================================
//
//	RenderLine
//
// 	Clips the given segment and adds any visible pieces to the line list.
//
//==========================================================================

#ifdef TEST_LIGHTS
static TCvarF	r_vlight0("r_vlight0", "0");
static TCvarF	r_vlight1("r_vlight1", "0");
static TCvarF	r_vlight2("r_vlight2", "0");
static TCvarF	r_vlight3("r_vlight3", "0");
#endif

static void RenderLine(seg_t* line, int clipflags)
{
	if (!line->linedef)
	{
		//	Miniseg
		return;
	}

	float dist = DotProduct(vieworg, line->normal) - line->dist;
	if (dist <= 0)
	{
	    //	Viewer is in back side or on plane
		return;
	}

    backsector = line->backsector;

	draw_vertex_t	wv[4];
	side_t*			sidedef;
	line_t*			linedef;
	float			wallh1;
	float			wallh2;
	float			hdelta;
	float			clipz1;
	float			clipz2;
	float			offshdelta;
	float			t_org;

    sidedef = line->sidedef;
    linedef = line->linedef;

#ifdef NEED_AXIS
	r_saxis = Normalize(*line->v2 - *line->v1);
	r_taxis = TVec(0, 0, -1);
	r_texorg = *line->v1 - r_saxis * (sidedef->textureoffset + line->offset);
#endif

#ifdef TEST_LIGHTS
	CalcLineLights(linedef, r_vlight0, r_vlight1, r_vlight2, r_vlight3);
#endif

	//FIXME this marks all lines
    // mark the segment as visible for auto map
    linedef->flags |= ML_MAPPED;

	wv[0].x = wv[1].x = line->v1->x;
	wv[0].y = wv[1].y = line->v1->y;

	wv[2].x = wv[3].x = line->v2->x;
	wv[2].y = wv[3].y = line->v2->y;

	wv[0].s = wv[1].s = sidedef->textureoffset + line->offset;
	wv[2].s = wv[3].s = sidedef->textureoffset + line->offset + line->length;

	float lights[4];
	lights[0] = r_region->params->lightlevel + line->vlight[0];
	lights[1] = r_region->params->lightlevel + line->vlight[1];
	lights[2] = r_region->params->lightlevel + line->vlight[2];
	lights[3] = r_region->params->lightlevel + line->vlight[3];

	float topz1 = r_ceiling->GetPointZ(line->v1->x, line->v1->y);
	float topz2 = r_ceiling->GetPointZ(line->v2->x, line->v2->y);
	float botz1 = r_floor->GetPointZ(line->v1->x, line->v1->y);
	float botz2 = r_floor->GetPointZ(line->v2->x, line->v2->y);

    if (!backsector)
    {
        // single sided line
		wv[0].z = botz1;
		wv[1].z = topz1;
		wv[2].z = topz2;
		wv[3].z = botz2;

		wallh1 = topz1 - botz1;
		wallh2 = topz2 - botz2;
		hdelta = topz2 - topz1;
		offshdelta = hdelta * line->offset / line->length;

        if (linedef->flags & ML_DONTPEGBOTTOM)
        {
            //	bottom of texture at bottom
			t_org = textureheight[sidedef->midtexture] -
				(botz1 > botz2 ? wallh2 - hdelta : wallh1);
        }
        else
		{
            // top of texture at top
			t_org = 0;
		}
		t_org += sidedef->rowoffset - offshdelta;

		wv[0].t = t_org + wallh1;
		wv[1].t = t_org;
		wv[2].t = t_org - hdelta;
		wv[3].t = t_org - hdelta + wallh2;

		wv[0].l = lights[0] + r_floor->light;
		wv[1].l = lights[1] + r_ceiling->light;
		wv[2].l = lights[2] + r_ceiling->light;
		wv[3].l = lights[3] + r_floor->light;

#ifdef NEED_AXIS
		r_texorg.z = topz1 - t_org;
#endif

		if (wallh1 || wallh2)
		{
			Drawer->DrawPolygon(wv, 4, sidedef->midtexture, clipflags, line->normal, r_use_fog);
		}

       	if (frontsector->ceiling.pic == skyflatnum &&
       		frontsector->ceilingheight != frontsector->skyheight)
		{
			wv[0].z = topz1;
			wv[1].z = wv[2].z = frontsector->skyheight;
			wv[3].z = topz2;
			R_RenderSkyPoly(wv, 4, clipflags);
		}
	}
    else
    {
		// two sided line

		r_back_floor = &backsector->floor;
		r_back_ceiling = &backsector->ceiling;

		wv[0].l = lights[0];
		wv[1].l = lights[1];
		wv[2].l = lights[2];
		wv[3].l = lights[3];

		float back_topz1 = r_back_ceiling->GetPointZ(line->v1->x, line->v1->y);
		float back_topz2 = r_back_ceiling->GetPointZ(line->v2->x, line->v2->y);
		float back_botz1 = r_back_floor->GetPointZ(line->v1->x, line->v1->y);
		float back_botz2 = r_back_floor->GetPointZ(line->v2->x, line->v2->y);

        // hack to allow height changes in outdoor areas
		if (r_ceiling->pic == skyflatnum &&
			r_back_ceiling->pic == skyflatnum)
		{
			topz1 = back_topz1;
			topz2 = back_topz2;
		}

		if (MAX(back_topz1, botz1) < topz1 || MAX(back_topz2, botz2) < topz2)
        {
			// top wall
			wv[0].z = MAX(back_topz1, botz1);
			wv[1].z = topz1;
			wv[2].z = topz2;
			wv[3].z = MAX(back_topz2, botz2);

			wallh1 = topz1 - back_topz1;
			wallh2 = topz2 - back_topz2;
			hdelta = topz2 - topz1;
			offshdelta = hdelta * line->offset / line->length;
			if (back_topz1 < botz1)
			{
				clipz1 = back_topz1 - botz1;
				clipz2 = back_topz2 - botz2;
			}
			else
			{
				clipz1 = 0;
				clipz2 = 0;
			}

            if (linedef->flags & ML_DONTPEGTOP)
            {
                // top of texture at top
				t_org = 0;
            }
            else
            {
                // bottom of texture
				t_org = textureheight[sidedef->toptexture] - wallh1;
            }
			t_org += sidedef->rowoffset - offshdelta;

			wv[0].t = t_org + wallh1 + clipz1;
			wv[1].t = t_org;
			wv[2].t = t_org - hdelta;
			wv[3].t = t_org - hdelta + wallh2 + clipz2;

#ifdef NEED_AXIS
			r_texorg.z = topz1 - t_org;
#endif

			Drawer->DrawPolygon(wv, 4, sidedef->toptexture, clipflags, line->normal, r_use_fog);
		}

		if (r_ceiling->pic == skyflatnum &&
			r_back_ceiling->pic != skyflatnum &&
			frontsector->ceilingheight != frontsector->skyheight)
		{
			wv[0].z = topz1;
			wv[1].z = wv[2].z = frontsector->skyheight;
			wv[3].z = topz2;
			R_RenderSkyPoly(wv, 4, clipflags);
		}

		if (MIN(back_botz1, topz1) > botz1 || MIN(back_botz2, topz2) > botz2)
        {
			// bottom wall
			wv[0].z = botz1;
			wv[1].z = MIN(back_botz1, topz1);
			wv[2].z = MIN(back_botz2, topz2);
			wv[3].z = botz2;

			wallh1 = back_botz1 - botz1;
			wallh2 = back_botz2 - botz2;
			hdelta = back_botz2 - back_botz1;
			offshdelta = hdelta * line->offset / line->length;
			if (back_botz1 > topz1)
			{
				clipz1 = back_botz1 - topz1;
				clipz2 = back_botz2 - topz2;
			}
			else
			{
				clipz1 = 0;
				clipz2 = 0;
			}

			if (linedef->flags & ML_DONTPEGBOTTOM)
            {
                // bottom of texture at bottom
                // top of texture at top
				t_org = topz1 - back_botz1;
            }
            else
            {
            	// top of texture at top
				t_org = 0;
            }
			t_org += sidedef->rowoffset - offshdelta;

			wv[0].t = t_org + wallh1;
			wv[1].t = t_org + clipz1;
			wv[2].t = t_org - hdelta + clipz2;
			wv[3].t = t_org - hdelta + wallh2;

#ifdef NEED_AXIS
			r_texorg.z = botz1 - (t_org + wallh1);
#endif

			Drawer->DrawPolygon(wv, 4, sidedef->bottomtexture, clipflags, line->normal, r_use_fog);
        }

		float midtopz1 = MIN(topz1, back_topz1);
		float midtopz2 = MIN(topz2, back_topz2);
		float midbotz1 = MAX(botz1, back_botz1);
		float midbotz2 = MAX(botz2, back_botz2);

		if (sidedef->midtexture &&
			(midtopz1 > midbotz1 || midtopz1 > midbotz1))
		{
			// masked midtexture
			float texh = textureheight[sidedef->midtexture];
			hdelta = midtopz2 - midtopz1;
			offshdelta = hdelta * line->offset / line->length;

			float			z_org;

			if (linedef->flags & ML_DONTPEGBOTTOM)
			{
				// bottom of texture at bottom
				// top of texture at top
				z_org = MIN(midbotz1, midbotz2) + texh;
			}
			else
			{
				// top of texture at top
				z_org = MAX(midtopz1, midtopz2);
			}
			z_org += sidedef->rowoffset - offshdelta;

			wv[0].z = z_org - texh;
			wv[1].z = z_org;
			wv[2].z = z_org;
			wv[3].z = z_org - texh;

			wv[0].t = texh;
			wv[1].t = 0;
			wv[2].t = 0;
			wv[3].t = texh;

			if (wv[0].z < midbotz1)
			{
				wv[0].t += wv[0].z - midbotz1;
				wv[0].z = midbotz1;
			}
			if (wv[1].z > midtopz1)
			{
				wv[1].t += wv[1].z - midtopz1;
				wv[1].z = midtopz1;
			}
			if (wv[2].z > midtopz2)
			{
				wv[2].t += wv[2].z - midtopz2;
				wv[2].z = midtopz2;
			}
			if (wv[3].z < midbotz2)
			{
				wv[3].t += wv[3].z - midbotz2;
				wv[3].z = midbotz2;
			}
			if (linedef->translucency)
			{
				R_DrawTranslucentPoly(wv, 4, dist, sidedef->midtexture,
					line->normal, linedef->translucency, 0, false);
			}
			else
			{
				Drawer->DrawMaskedPolygon(wv, 4, sidedef->midtexture,
					0, 15, line->normal, r_use_fog);
			}
		}

		sec_region_t *reg;
		for (reg = backsector->botregion; reg->next; reg = reg->next)
		{
			TPlane *extratop = reg->next->floor;
			TPlane *extrabot = reg->ceiling;
			side_t *extraside = &cl_level.sides[reg->extraline->sidenum[0]];

			float extratopz1 = extratop->GetPointZ(line->v1->x, line->v1->y);
			float extratopz2 = extratop->GetPointZ(line->v2->x, line->v2->y);
			float extrabotz1 = extrabot->GetPointZ(line->v1->x, line->v1->y);
			float extrabotz2 = extrabot->GetPointZ(line->v2->x, line->v2->y);

			midtopz1 = MIN(extratopz1, topz1);
			midtopz2 = MIN(extratopz2, topz2);
			midbotz1 = MAX(extrabotz1, botz1);
			midbotz2 = MAX(extrabotz2, botz2);
			if (midtopz1 - midbotz1 > 0.5 || midtopz2 - midbotz2 > 0.5)
			{
				wv[0].z = midbotz1;
				wv[1].z = midtopz1;
				wv[2].z = midtopz2;
				wv[3].z = midbotz2;

				wallh1 = extratopz1 - midbotz1;
				wallh2 = extratopz2 - midbotz2;
				hdelta = extratopz2 - extratopz1;
				offshdelta = hdelta * line->offset / line->length;
				if (extratopz1 > topz1)
				{
					clipz1 = extratopz1 - topz1;
					clipz2 = extratopz2 - topz2;
				}
				else
				{
					clipz1 = 0;
					clipz2 = 0;
				}

				t_org = 0;
				t_org += sidedef->rowoffset - offshdelta;

				wv[0].t = t_org + wallh1;
				wv[1].t = t_org + clipz1;
				wv[2].t = t_org - hdelta + clipz2;
				wv[3].t = t_org - hdelta + wallh2;

#ifdef NEED_AXIS
				r_texorg.z = midbotz1 - (t_org + wallh1);
#endif

				Drawer->DrawPolygon(wv, 4, extraside->midtexture, clipflags, line->normal, r_use_fog);
			}
		}
    }
}

//==========================================================================
//
//	RenderSecPlane
//
//==========================================================================

static void	RenderSecPlane(const sec_plane_t &plane, const seg_t *line,
	int count, int vlindex, int clipflags)
{
	if (!plane.pic)
	{
		return;
	}

	float dist = DotProduct(vieworg, plane.normal) - plane.dist;
	if (dist <= 0)
	{
	    //	Viewer is in back side or on plane
		return;
	}

	if (plane.angle)
	{
		r_saxis = TVec(mcos(plane.angle), msin(plane.angle), 0);
		r_taxis = TVec(msin(plane.angle), -mcos(plane.angle), 0);
		r_texorg = -(plane.xoffs * r_saxis + plane.yoffs * r_taxis);
	}
	else
	{
		r_saxis = TVec(1, 0, 0);
		r_taxis = TVec(0, -1, 0);
		r_texorg = TVec(-plane.xoffs, plane.yoffs, 0);
	}

	r_saxis *= plane.xscale;
	r_taxis *= plane.yscale;

#ifdef __GNUC__
	draw_vertex_t tmpv[count];
#else
	draw_vertex_t *tmpv = new draw_vertex_t[count];
#endif

	const seg_t *prevline = line + (count - 1);
   	for (int i = 0; i < count; i++)
    {
		TVec &v = *line[i].v1;
		draw_vertex_t &dst = tmpv[vlindex ? count - i - 1 : i];
		dst.x = v.x;
		dst.y = v.y;
		dst.z = plane.GetPointZ(dst.x, dst.y);
		dst.s = DotProduct(dst - r_texorg, r_saxis);
		dst.t = DotProduct(dst - r_texorg, r_taxis);
		dst.l = r_region->params->lightlevel + plane.light +
			(line[i].vlight[vlindex] + prevline->vlight[3 - vlindex]) / 2;
		prevline = &line[i];
	}

	if (plane.pic == skyflatnum)
	{
		R_RenderSkyPoly(tmpv, count, clipflags);
	}
	else if (plane.translucency)
	{
		R_DrawTranslucentPoly(tmpv, count, dist, plane.pic,
			plane.normal, plane.translucency, 0, false);
	}
	else
	{
		Drawer->DrawPolygon(tmpv, count, plane.pic, clipflags, plane.normal, r_use_fog);
	}
#ifndef __GNUC__
	delete tmpv;
#endif
}

//==========================================================================
//
//	RenderRegion
//
// 	Determine floor/ceiling planes.
// 	Draw one or more line segments.
//
//==========================================================================

static void RenderRegion(subsector_t *sub, sec_region_t *region, int clipflags)
{
    int				count;
    seg_t*			line;
	int 			polyCount;
	seg_t**			polySeg;

	r_region = region;
	r_floor = region->floor;
	r_ceiling = region->ceiling;

	if (sub->poly && !r_back2front)
	{
		//	Render the polyobj in the subsector first
		polyCount = sub->poly->numsegs;
		polySeg = sub->poly->segs;
		while (polyCount--)
		{
			RenderLine(*polySeg++, clipflags);
		}
	}

    count = sub->numlines;
    line = &cl_level.segs[sub->firstline];
    while (count--)
    {
		RenderLine(line, clipflags);
		line++;
    }

   	count = sub->numlines;
    line = &cl_level.segs[sub->firstline];

	RenderSecPlane(*r_floor, line, count, 0, clipflags);
	if (r_ceiling->pic != skyflatnum)
	{
		RenderSecPlane(*r_ceiling, line, count, 1, clipflags);
	}
	else
	{
		sec_plane_t	r_sky;

		r_sky.Set(TVec(0, 0, -1), -frontsector->skyheight);
		r_sky.pic = skyflatnum;

		RenderSecPlane(r_sky, line, count, 1, clipflags);
	}

	if (sub->poly && r_back2front)
	{
		//	Render the polyobj
		polyCount = sub->poly->numsegs;
		polySeg = sub->poly->segs;
		while (polyCount--)
		{
			RenderLine(*polySeg++, clipflags);
		}
	}
}

//==========================================================================
//
//	RenderSubsector
//
//==========================================================================

static void RenderSubsector(int num, int clipflags)
{
    subsector_t*	sub;
	sec_region_t*	region;
	sec_region_t*	startregion;

//FIXME do this in node loading
#ifdef PARANOID
    if (num >= cl_level.numsubsectors)
		I_Error("RenderSubsector: ss %i with numss = %i", num, cl_level.numsubsectors);
#endif

    sub = &cl_level.subsectors[num];
    frontsector = sub->sector;

	if (sub->visframe != r_visframecount)
 	{
		return;
	}

	if (!frontsector->linecount)
	{
		//	Skip sectors containing original polyobjs
		return;
	}

	if (!frontsector->botregion->next)
	{
		//	Just one region
		RenderRegion(sub, frontsector->botregion, clipflags);
		return;
	}

	startregion = frontsector->botregion;
	while (startregion->next)
	{
		if (DotProduct(vieworg, startregion->ceiling->normal) -
			startregion->ceiling->dist >= 0)
		{
			break;
		}
		startregion = startregion->next;
	}

	if (r_back2front)
	{
		for (region = frontsector->topregion; region != startregion; region = region->prev)
		{
			RenderRegion(sub, region, clipflags);
		}
		for (region = frontsector->botregion; region != startregion; region = region->next)
		{
			RenderRegion(sub, region, clipflags);
		}
		RenderRegion(sub, startregion, clipflags);
	}
	else
	{
		for (region = startregion; region; region = region->prev)
		{
			RenderRegion(sub, region, clipflags);
		}
		for (region = startregion->next; region; region = region->next)
		{
			RenderRegion(sub, region, clipflags);
		}
	}
}

//==========================================================================
//
//	RenderBSPNode
//
// 	Renders all subsectors below a given node, traversing subtree
// recursively. Just call with BSP root.
//
//==========================================================================

static void RenderBSPNode(int bspnum, float *bbox, int clipflags)
{
	// cull the clipping planes if not trivial accept
	if (clipflags)
	{
		for (int i = 0; i < 4; i++)
		{
			if (!(clipflags & view_clipplanes[i].clipflag))
				continue;	// don't need to clip against it

			// generate accept and reject points

			int *pindex = r_frustum_indexes[i];

			TVec rejectpt;

			rejectpt[0] = bbox[pindex[0]];
			rejectpt[1] = bbox[pindex[1]];
			rejectpt[2] = bbox[pindex[2]];

			float d;

			d = DotProduct(rejectpt, view_clipplanes[i].normal);
			d -= view_clipplanes[i].dist;

			if (d <= 0)
				return;

			TVec acceptpt;

			acceptpt[0] = bbox[pindex[3+0]];
			acceptpt[1] = bbox[pindex[3+1]];
			acceptpt[2] = bbox[pindex[3+2]];

			d = DotProduct(acceptpt, view_clipplanes[i].normal);
			d -= view_clipplanes[i].dist;

			if (d >= 0)
				clipflags ^= view_clipplanes[i].clipflag;	// node is entirely on screen
		}
	}

    // Found a subsector?
    if (bspnum & NF_SUBSECTOR)
    {
		if (bspnum == -1)
		    RenderSubsector(0, clipflags);
		else
		    RenderSubsector(bspnum & (~NF_SUBSECTOR), clipflags);
		return;
    }
		
	node_t* bsp = &cl_level.nodes[bspnum];
    
	if (bsp->visframe != r_visframecount)
 	{
		return;
	}

    // Decide which side the view point is on.
	int side = bsp->PointOnSide(vieworg);

	if (r_back2front)
	{
    	// Divide back space.
		RenderBSPNode(bsp->children[side ^ 1], bsp->bbox[side ^ 1], clipflags);

    	// Recursively divide front space.
	    RenderBSPNode(bsp->children[side], bsp->bbox[side], clipflags);
	}
	else
	{
    	// Recursively divide front space.
	    RenderBSPNode(bsp->children[side], bsp->bbox[side], clipflags);

    	// Divide back space.
		RenderBSPNode(bsp->children[side ^ 1], bsp->bbox[side ^ 1], clipflags);
	}
}

//==========================================================================
//
//	R_RenderWorld
//
//==========================================================================

void R_RenderWorld(void)
{
	float	dummy_bbox[6] = {-99999, -99999, -99999, 99999, 9999, 99999};

	R_SetUpFrustumIndexes();

	UpdateNodeBBox(cl_level.numnodes - 1, dummy_bbox);

	RenderBSPNode(cl_level.numnodes - 1, dummy_bbox, 15);	// head node is the last node output

	//	In software actual drawing is done here
	Drawer->WorldDrawing();
}


