//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2001 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Sky rendering. The DOOM sky is a texture map like any wall, wrapping
//**  around. A 1024 columns equal 360 degrees. The default sky map is 256
//**  columns and repeats 4 times on a 320 screen
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "r_local.h"

// MACROS ------------------------------------------------------------------

#define LIGHTNING_SPECIAL 	198
#define LIGHTNING_SPECIAL2 	199
#define SKYCHANGE_SPECIAL 	200

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void R_LightningFlash(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int 			Sky1BaseTexture;
static int 			Sky2BaseTexture;
static int 			Sky1TopBaseTexture;
static int 			Sky2TopBaseTexture;
static int 			Sky1BotBaseTexture;
static int 			Sky2BotBaseTexture;

static int 			Sky1Texture;
static int 			Sky2Texture;
static int 			Sky1TopTexture;
static int 			Sky2TopTexture;
static int 			Sky1BotTexture;
static int 			Sky2BotTexture;

static float		Sky1ColumnOffset;
static float		Sky2ColumnOffset;
static float		Sky1ScrollDelta;
static float		Sky2ScrollDelta;

static float		Sky1TopXOffset;
static float		Sky1TopYOffset;
static float		Sky2TopXOffset;
static float		Sky2TopYOffset;
static float		Sky1TopXDelta;
static float		Sky1TopYDelta;
static float		Sky2TopXDelta;
static float		Sky2TopYDelta;

static boolean 		DoubleSky;

static boolean		LevelHasLightning;
static int			NextLightningFlash;
static int			LightningFlash;
static int			*LightningLightLevels;

static TClipPlane	sky_clip[6][4];

static float		skytop;
static float		skybot;

static draw_vertex_t	*p_vert;
static int				r_emited;
static int				max_verts;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	R_InitSky
//
//	Called at level load.
//
//==========================================================================

void R_InitSky(const mapInfo_t &info)
{
	Sky1Texture = Sky1BaseTexture = info.sky1Texture;
	Sky2Texture = Sky2BaseTexture = info.sky2Texture;
	Sky1ScrollDelta = info.sky1ScrollDelta;
	Sky2ScrollDelta = info.sky2ScrollDelta;
	Sky1ColumnOffset = 0.0;
	Sky2ColumnOffset = 0.0;
	DoubleSky =	info.doubleSky;

	Sky1TopTexture = Sky1TopBaseTexture = Sky1Texture;
	Sky2TopTexture = Sky2TopBaseTexture = Sky2Texture;
	Sky1BotTexture = Sky1BotBaseTexture = Sky1Texture;
	Sky2BotTexture = Sky2BotBaseTexture = Sky2Texture;
	Sky1TopXOffset = 0;
	Sky1TopYOffset = 0;
	Sky2TopXOffset = 0;
	Sky2TopYOffset = 0;
	Sky1TopXDelta = 32;
	Sky1TopYDelta = 8;
	Sky2TopXDelta = 32;
	Sky2TopYDelta = 8;

	// Check if the level is a lightning level
	LevelHasLightning = false;
	LightningFlash = 0;
	if (info.lightning)
	{
		int secCount = 0;
		for (int i = 0; i < cl_level.numsectors; i++)
		{
			if (cl_level.sectors[i].ceiling.pic == skyflatnum ||
				cl_level.sectors[i].special == LIGHTNING_SPECIAL ||
				cl_level.sectors[i].special == LIGHTNING_SPECIAL2)
			{
				secCount++;
			}
		}
		if (secCount)
		{
			LevelHasLightning = true;
			LightningLightLevels = (int *)Z_Malloc(secCount * sizeof(int),
				PU_LEVEL, NULL);
			NextLightningFlash = ((rand() & 15) + 5) * 35; // don't flash at level start
		}
	}

	float skyheight = textureheight[Sky1Texture];
	float topwidth = 128.0;
	if (skyheight <= 128.0)
	{
		skytop = 95;
	}
	else
	{
		skytop = 190;
	}
	skybot = skytop - skyheight;

//	if (DoubleSky)
//	{
//		topwidth = 8 * 1024;
//	}

	sky_clip[0][0].Set(Normalize(TVec(128, -128, 0)), 0);
	sky_clip[0][1].Set(Normalize(TVec(128, 128, 0)), 0);
	sky_clip[0][2].Set(Normalize(TVec(skytop, 0, -128)), 0);
	sky_clip[0][3].Set(Normalize(TVec(-skybot, 0, 128)), 0);

	sky_clip[1][0].Set(Normalize(TVec(-128, -128, 0)), 0);
	sky_clip[1][1].Set(Normalize(TVec(128, -128, 0)), 0);
	sky_clip[1][2].Set(Normalize(TVec(0, -skytop, -128)), 0);
	sky_clip[1][3].Set(Normalize(TVec(0, skybot, 128)), 0);

	sky_clip[2][0].Set(Normalize(TVec(-128, 128, 0)), 0);
	sky_clip[2][1].Set(Normalize(TVec(-128, -128, 0)), 0);
	sky_clip[2][2].Set(Normalize(TVec(-skytop, 0, -128)), 0);
	sky_clip[2][3].Set(Normalize(TVec(skybot, 0, 128)), 0);

	sky_clip[3][0].Set(Normalize(TVec(128, 128, 0)), 0);
	sky_clip[3][1].Set(Normalize(TVec(-128, 128, 0)), 0);
	sky_clip[3][2].Set(Normalize(TVec(0, skytop, -128)), 0);
	sky_clip[3][3].Set(Normalize(TVec(0, -skybot, 128)), 0);

	sky_clip[4][0].Set(Normalize(TVec(0, -skytop, topwidth)), 0);
	sky_clip[4][1].Set(Normalize(TVec(0, skytop, topwidth)), 0);
	sky_clip[4][2].Set(Normalize(TVec(skytop, 0, topwidth)), 0);
	sky_clip[4][3].Set(Normalize(TVec(-skytop, 0, topwidth)), 0);

	sky_clip[5][0].Set(Normalize(TVec(0, skybot, -128)), 0);
	sky_clip[5][1].Set(Normalize(TVec(0, -skybot, -128)), 0);
	sky_clip[5][2].Set(Normalize(TVec(-skybot, 0, -128)), 0);
	sky_clip[5][3].Set(Normalize(TVec(skybot, 0, -128)), 0);

	for (int j = 0; j < 6; j++)
	{
		for (int i = 0; i < 4; i++)
		{
			sky_clip[j][i].next = i == 3 ? NULL : &sky_clip[j][i + 1];
			sky_clip[j][i].clipflag = 1 << i;
		}
	}

	//	Precache textures
	Drawer->SetSkyTexture(Sky1Texture, DoubleSky);
	Drawer->SetSkyTexture(Sky2Texture, false);
	Drawer->SetSkyTexture(Sky1TopTexture, DoubleSky);
	Drawer->SetSkyTexture(Sky2TopTexture, false);
	Drawer->SetSkyTexture(Sky1BotTexture, DoubleSky);
	Drawer->SetSkyTexture(Sky2BotTexture, false);
}

//==========================================================================
//
//	R_AnimateSky
//
//==========================================================================

void R_AnimateSky(void)
{
	//	Update sky column offsets
	Sky1ColumnOffset += Sky1ScrollDelta * host_frametime;
	Sky2ColumnOffset += Sky2ScrollDelta * host_frametime;
	Sky1TopXOffset += Sky1TopXDelta * host_frametime;
	Sky1TopYOffset += Sky1TopYDelta * host_frametime;
	Sky2TopXOffset += Sky2TopXDelta * host_frametime;
	Sky2TopYOffset += Sky2TopYDelta * host_frametime;

	//	Update lightning
	if (LevelHasLightning)
	{
		if (!NextLightningFlash || LightningFlash)
		{
			R_LightningFlash();
		}
		else
		{
			NextLightningFlash--;
		}
	}
}

//==========================================================================
//
//	R_LightningFlash
//
//==========================================================================

static void R_LightningFlash(void)
{
	int 		i;
	sector_t 	*tempSec;
	int 		*tempLight;
	boolean 	foundSec;
	int 		flashLight;

	if (LightningFlash)
	{
		LightningFlash--;
		if (LightningFlash)
		{
			tempLight = LightningLightLevels;
			tempSec = cl_level.sectors;
			for (i = 0; i < cl_level.numsectors; i++, tempSec++)
			{
				if (tempSec->ceiling.pic == skyflatnum ||
					tempSec->special == LIGHTNING_SPECIAL ||
					tempSec->special == LIGHTNING_SPECIAL2)
				{
					if (*tempLight < tempSec->params.lightlevel - 4)
					{
						tempSec->params.lightlevel -= 4;
					}
					tempLight++;
				}
			}
		}					
		else
		{ // remove the alternate lightning flash special
			tempLight = LightningLightLevels;
			tempSec = cl_level.sectors;
			for (i = 0; i < cl_level.numsectors; i++, tempSec++)
			{
				if (tempSec->ceiling.pic == skyflatnum
					|| tempSec->special == LIGHTNING_SPECIAL
					|| tempSec->special == LIGHTNING_SPECIAL2)
				{
					tempSec->params.lightlevel = *tempLight;
					tempLight++;
				}
			}
			Sky1Texture = Sky1BaseTexture;
			Sky1TopTexture = Sky1TopBaseTexture;
			Sky1BotTexture = Sky1BotBaseTexture;
		}
		return;
	}

	LightningFlash = (rand() & 7) + 8;
	flashLight = 200 + (rand() & 31);
	tempSec = cl_level.sectors;
	tempLight = LightningLightLevels;
	foundSec = false;
	for (i = 0; i < cl_level.numsectors; i++, tempSec++)
	{
		if (tempSec->ceiling.pic == skyflatnum ||
			tempSec->special == LIGHTNING_SPECIAL ||
			tempSec->special == LIGHTNING_SPECIAL2)
		{
			*tempLight = tempSec->params.lightlevel;
			if (tempSec->special == LIGHTNING_SPECIAL)
			{ 
				tempSec->params.lightlevel += 64;
				if (tempSec->params.lightlevel > flashLight)
				{
					tempSec->params.lightlevel = flashLight;
				}
			}
			else if (tempSec->special == LIGHTNING_SPECIAL2)
			{
				tempSec->params.lightlevel += 32;
				if (tempSec->params.lightlevel > flashLight)
				{
					tempSec->params.lightlevel = flashLight;
				}
			}
			else
			{
				tempSec->params.lightlevel = flashLight;
			}
			if (tempSec->params.lightlevel < *tempLight)
			{
				tempSec->params.lightlevel = *tempLight;
			}
			tempLight++;
			foundSec = true;
		}
	}
	if (foundSec)
	{
		Sky1Texture = Sky2BaseTexture; // set alternate sky
		Sky1TopTexture = Sky2TopBaseTexture; // set alternate sky
		Sky1BotTexture = Sky2BotBaseTexture; // set alternate sky
		S_StartSoundName("ThunderCrash");
	}
	// Calculate the next lighting flash
	if (!NextLightningFlash)
	{
		if ((rand() & 0xff) < 50)
		{
			// Immediate Quick flash
			NextLightningFlash = (rand() & 15) + 16;
		}
		else
		{
			if ((rand() & 0xff) < 128 && !(cl_level.tictime & 32))
			{
				NextLightningFlash = ((rand() & 7) + 2) * 35;
			}
			else
			{
				NextLightningFlash = ((rand() & 15) + 5) * 35;
			}
		}
	}
}

//==========================================================================
//
//	R_ForceLightning
//
//==========================================================================

void R_ForceLightning(void)
{
	NextLightningFlash = 0;
}

//==========================================================================
//
//	ClipSkyEdge
//
//==========================================================================

static void ClipSkyEdge(const draw_vertex_t &v0, const draw_vertex_t &v1,
	TClipPlane *clip, TClipPlane *baseclip, int clipflags)
{
	if (clip)
	{
		do
		{
			if (!(clipflags & clip->clipflag))
			{
				continue;
			}

			float d0 = DotProduct(v0, clip->normal) - clip->dist;
			float d1 = DotProduct(v1, clip->normal) - clip->dist;
			if (d0 >= 0)
			{
				// point 0 is unclipped
				if (d1 >= 0)
				{
					// both points are unclipped
					continue;
				}

				// only point 1 is clipped
				clip->entered = true;

				float f = d0 / (d0 - d1);
				clip->enter = v0 + f * (v1 - v0);

				ClipSkyEdge(v0, clip->enter, clip->next, baseclip, clipflags);
				if (clip->exited)
				{
					clip->entered = false;
					clip->exited = false;
					ClipSkyEdge(clip->enter, clip->exit, baseclip, baseclip,
						clipflags ^ clip->clipflag);
				}
				return;
			}
			else
			{
				// point 0 is clipped

				if (d1 < 0)
				{
					// both points are clipped
					return;
				}

				// only point 0 is clipped
				clip->exited = true;

				float f = d0 / (d0 - d1);
				clip->exit = v0 + f * (v1 - v0);

				if (clip->entered)
				{
					clip->entered = false;
					clip->exited = false;
					ClipSkyEdge(clip->enter, clip->exit, baseclip, baseclip,
						clipflags ^ clip->clipflag);
				}
				ClipSkyEdge(clip->exit, v1, clip->next, baseclip, clipflags);
				return;
			}
		} while ((clip = clip->next) != NULL);
	}

	if (r_emited < max_verts)
	{
		//	add the vertex
		p_vert[r_emited++] = v1;
	}
}

//==========================================================================
//
//	R_RenderSkyPoly
//
//==========================================================================

void R_RenderSkyPoly(draw_vertex_t *cv, int count, int clipflags)
{
	int		i;

	Drawer->DrawZPolygon(cv, count, clipflags);

	//	Clip polygon against view clipplanes
	for (i = 0; i < 4; i++)
	{
		view_clipplanes[i].entered = false;
		view_clipplanes[i].exited = false;
	}

#ifdef __GNUC__
	draw_vertex_t dv[count + 16];
#else
	draw_vertex_t *dv = new draw_vertex_t[count + 16];
#endif
	p_vert = dv;
	max_verts = count + 16;

	r_emited = 0;
	for (i = 0; i < count; i++)
	{
		ClipSkyEdge(cv[i ? i - 1 : count - 1], cv[i], view_clipplanes,
			view_clipplanes, clipflags);
	}

	if (!r_emited)
	{
#ifndef __GNUC__
		delete dv;
#endif
		return;
	}

	//	Translate to camera space
	int newcount = r_emited;

	for (i = 0; i < newcount; i++)
	{
		dv[i].x	-= vieworg.x;
		dv[i].y	-= vieworg.y;
		dv[i].z	-= vieworg.z;
	}

	for (int j = 0; j < 6; j++)
	{
		for (i = 0; i < 4; i++)
		{
			sky_clip[j][i].entered = false;
			sky_clip[j][i].exited = false;
		}
	}
#ifdef __GNUC__
	draw_vertex_t pv[newcount + 4];
#else
	draw_vertex_t *pv = new draw_vertex_t[newcount + 4];
#endif
	p_vert = pv;

	//	Top
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[4],
			sky_clip[4], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = skytop / v.z;
			float x = v.y * z1;
			float y = v.x * z1;

			v.x = y;
			v.y = x;
			v.z = skytop;

			v.s = -x + 128;
			v.t = y + 128;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front && Sky2TopTexture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2TopTexture, Sky2TopXOffset, Sky2TopYOffset, false);
			}
			if (Sky1TopTexture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky1TopTexture, Sky1TopXOffset, Sky1TopYOffset, true);
			}
			if (!r_back2front && Sky2TopTexture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2TopTexture, Sky2TopXOffset, Sky2TopYOffset, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1TopTexture, Sky1TopXOffset, Sky1TopYOffset, false);
		}
	}

	//	Bottom
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[5],
			sky_clip[5], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = skybot / v.z;
			float x = v.y * z1;
			float y = v.x * z1;

			v.x = y;
			v.y = x;
			v.z = skybot;

			v.s = -x + 128;
			v.t = y + 128;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2BotTexture, 0, 0, false);
			}
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1BotTexture, 0, 0, true);
			if (!r_back2front)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2BotTexture, 0, 0, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1BotTexture, 0, 0, false);
		}
	}

	//	Side 1
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[0],
			sky_clip[0], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = 128.0f / v.x;
			float x = v.y * z1;
			float y = v.z * z1;

			v.x = 128;
			v.y = x;
			v.z = y;

			v.s = -x + 128;
			v.t = -y + skytop;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, true);
			if (!r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, false);
		}
	}

	//	Side 2
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[1],
			sky_clip[1], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = -128.0f / v.y;
			float x = v.x * z1;
			float y = v.z * z1;

			v.x = x;
			v.y = -128;
			v.z = y;

			v.s = -x + 128 + 256;
			v.t = -y + skytop;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, true);
			if (!r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, false);
		}
	}

	//	Side 3
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[2],
			sky_clip[2], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = -128.0f / v.x;
			float x = v.y * z1;
			float y = v.z * z1;

			v.x = -128;
			v.y = x;
			v.z = y;

			v.s = x + 128 + 512;
			v.t = -y + skytop;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, true);
			if (!r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, false);
		}
	}

	//	Side 4
	r_emited = 0;
	for (i = 0; i < newcount; i++)
	{
		ClipSkyEdge(dv[i ? i - 1 : newcount - 1], dv[i], sky_clip[3],
			sky_clip[3], 15);
	}
	if (r_emited)
	{
		for (i = 0; i < r_emited; i++)
		{
			draw_vertex_t	&v = pv[i];

			float z1 = 128.0f / v.y;
			float x = v.x * z1;
			float y = v.z * z1;

			v.x = x;
			v.y = 128;
			v.z = y;

			v.s = x + 128 + 768;
			v.t = -y + skytop;
			v.l = 0;
		}
		if (DoubleSky)
		{
			if (r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, true);
			if (!r_back2front && Sky2Texture)
			{
				Drawer->DrawSkyPolygon(pv, r_emited, Sky2Texture, Sky2ColumnOffset, 0, false);
			}
		}
		else
		{
			Drawer->DrawSkyPolygon(pv, r_emited, Sky1Texture, Sky1ColumnOffset, 0, false);
		}
	}
#ifndef __GNUC__
	delete dv;
	delete pv;
#endif
}

