//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************
//**
//**	Creates PVS(Potentially Visible Set)
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "glvis.h"

// MACROS ------------------------------------------------------------------

//#define TEST_SPLIT

#define	ML_TWOSIDED			0x0004	// Backside will not be present at all

#define ON_EPSILON			0.1

#define set_vis(s1, s2)		vis[s1 * numsubsectors + s2] = true
#define get_vis(s1, s2)		vis[s1 * numsubsectors + s2]

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

#ifdef TEST_SPLIT
float				split_factor = 8.0;
#endif

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte			*vis;

static seg_t		*checks1;
static seg_t		*checks2;
static TPlane		side1;
static TPlane		side2;
static TVec			checkv11;
static TVec			checkv12;
static TVec			checkv21;
static TVec			checkv22;

static line_t		*enters1;
static line_t		*exits1;
static line_t		*enters2;
static line_t		*exits2;
static line_t		*mids;

static int			validcount;

// CODE --------------------------------------------------------------------

//==========================================================================
//
//	PlaneSide2
//
//	Returns side 1 (front), -1 (back), or 0 (on).
//
//==========================================================================

static int PlaneSide2(const TVec &point, const TPlane* plane)
{
//	float dot = DotProduct(point, plane->normal) - plane->dist;
	float dot = point.x * plane->normal.x + point.y * plane->normal.y - plane->dist;
    return dot < -ON_EPSILON ? -1 : dot > ON_EPSILON;
}

//==========================================================================
//
//	CheckLine
//
//==========================================================================

static bool	CheckLine(line_t *line)
{
	if (line->flags & ML_TWOSIDED)
	{
		//	Ignore two-sided lines
	    return true;
	}

	if (line->validcount == validcount)
	{
		//	Already checked
		return true;
	}
	line->validcount = validcount;

	int ss11 = PlaneSide2(*line->v1, checks1);
	int ss12 = PlaneSide2(*line->v2, checks1);

	if (ss11 <= 0 && ss12 <= 0)
	{
		//	Behind seg1
	    return true;
	}

	int ss21 = PlaneSide2(*line->v1, checks2);
	int ss22 = PlaneSide2(*line->v2, checks2);

	if (ss21 <= 0 && ss22 <= 0)
	{
		//	Behind seg2
	    return true;
	}

	int s11 = PlaneSide2(*line->v1, &side1);
	int s12 = PlaneSide2(*line->v2, &side1);

	if (s11 <= 0 && s12 <= 0)
	{
		//	Behind side1
	    return true;
	}

	int s21 = PlaneSide2(*line->v1, &side2);
	int s22 = PlaneSide2(*line->v2, &side2);

	if (s21 <= 0 && s22 <= 0)
	{
		//	Behind side2
	    return true;
	}

	//	Handle cases when line crosses both sides and a seg's line to make
	// sure that it crosses sides between segs
	if (ss11 <= 0 && s12 <= 0)
	{
		if (PlaneSide2(checkv12, line) >= 0)
		{
			return true;
		}
	}

	if (ss12 <= 0 && s11 <= 0)
	{
		if (PlaneSide2(checkv12, line) <= 0)
		{
			return true;
		}
	}

	if (ss12 <= 0 && s21 <= 0)
	{
		if (PlaneSide2(checkv11, line) >= 0)
		{
			return true;
		}
	}

	if (ss11 <= 0 && s22 <= 0)
	{
		if (PlaneSide2(checkv11, line) <= 0)
		{
			return true;
		}
	}

	if (ss21 <= 0 && s22 <= 0)
	{
		if (PlaneSide2(checkv22, line) >= 0)
		{
			return true;
		}
	}

	if (ss22 <= 0 && s21 <= 0)
	{
		if (PlaneSide2(checkv22, line) <= 0)
		{
			return true;
		}
	}

	if (ss22 <= 0 && s11 <= 0)
	{
		if (PlaneSide2(checkv21, line) >= 0)
		{
			return true;
		}
	}

	if (ss21 <= 0 && s12 <= 0)
	{
		if (PlaneSide2(checkv21, line) <= 0)
		{
			return true;
		}
	}

	if ((s11 <= 0 && s22 <= 0) || (s12 <= 0 && s21 <= 0))
	{
		//	Line entirely blocks sight
		return false;
	}

	if (s11 <= 0)
 	{
		line->next = enters1;
		enters1 = line;
 	}
 	else if (s12 <= 0)
	{
		line->next = exits1;
		exits1 = line;
	}
	else if (s21 <= 0)
 	{
		line->next = enters2;
		enters2 = line;
 	}
 	else if (s22 <= 0)
	{
		line->next = exits2;
		exits2 = line;
	}
	else
	{
		line->next = mids;
		mids = line;
	}

	return true;
}

//==========================================================================
//
//	CrossBSPNode
//
//	Returns true if strace crosses the given node successfully.
//
//==========================================================================

static bool CrossBSPNode(int bspnum)
{
    node_t*	bsp;
	int		num;
   	int		count;
    seg_t*	seg;

    if (bspnum & NF_SUBSECTOR)
    {
		if (bspnum == -1)
		    num = 0;
		else
		    num = bspnum & (~NF_SUBSECTOR);
    	count = subsectors[num].numlines;
	    seg = &segs[subsectors[num].firstline];
    	for ( ; count ; seg++, count--)
	    {
			if (seg->linedef && !CheckLine(seg->linedef))
			{
				return false;
			}
		}
	}
	else
	{
	    bsp = &nodes[bspnum];
		int s11 = PlaneSide2(checkv11, bsp);
		int s12 = PlaneSide2(checkv12, bsp);
		int s21 = PlaneSide2(checkv21, bsp);
		int s22 = PlaneSide2(checkv22, bsp);
		if (s11 > 0 || s12 > 0 || s21 > 0 || s22 > 0)
		{
			if (!CrossBSPNode(bsp->children[0]))
			{
				return false;
			}
		}
		if (s11 < 0 || s12 < 0 || s21 < 0 || s22 < 0)
		{
			if (!CrossBSPNode(bsp->children[1]))
			{
				return false;
			}
		}
	}
	return true;
}

//==========================================================================
//
//	CheckMultiLine
//
//==========================================================================

static bool CheckMultiLine(line_t *line, line_t *exits)
{
	line_t	*lp;

	for (lp = mids; lp; lp = lp->next)
	{
		if (lp->validcount == validcount)
		{
			//	Already used
			continue;
		}
		if (line->v2 == lp->v1)
		{
			lp->validcount = validcount;
			if (!CheckMultiLine(lp, exits))
			{
				return false;
			}
		}
	}

	for (lp = exits; lp; lp = lp->next)
	{
		if (line->v2 == lp->v1)
		{
			return false;
		}
	}

	return true;
}

//==========================================================================
//
//	CheckSegParts
//
//==========================================================================

static bool	CheckSegParts(void)
{
	if (checkv12 == checkv21)
	{
		side1.Set(checks1->normal, checks1->dist);
	}
	else
	{
		side1.Set2Points(checkv12, checkv21);
	}

	if (checkv22 == checkv11)
	{
		side2.Set(checks2->normal, checks2->dist);
	}
	else
	{
		side2.Set2Points(checkv22, checkv11);
	}

	enters1 = NULL;
	exits1 = NULL;
	enters2 = NULL;
	exits2 = NULL;
	mids = NULL;

	validcount++;
	if (!CrossBSPNode(numnodes - 1))
	{
		return false;
	}

	if (enters1 && exits2)
	{
		while (enters1)
		{
			validcount++;
			if (!CheckMultiLine(enters1, exits2))
			{
				return false;
			}
			enters1 = enters1->next;
		}
	}
	if (enters2 && exits1)
	{
		while (enters2)
		{
			validcount++;
			if (!CheckMultiLine(enters2, exits1))
			{
				return false;
			}
			enters2 = enters2->next;
		}
	}

	return true;
}

//==========================================================================
//
//	CheckSegs
//
//==========================================================================

static bool CheckSegs(seg_t *seg1, seg_t *seg2)
{
	int s11 = PlaneSide2(*seg1->v1, seg2);
	int s12 = PlaneSide2(*seg1->v2, seg2);
	int s21 = PlaneSide2(*seg2->v1, seg1);
	int s22 = PlaneSide2(*seg2->v2, seg1);

	if ((!s11 && !s12) || (!s21 && !s22))
	{
		//	Segs are on the same line
		return false;
	}

	//	Make segs facing each other
	if (!(s11 + s12))
	{
		if (s21 < 0 || s22 < 0)
		{
			seg1 = seg1->partner;
		}
		if (s12 < 0)
		{
			seg2 = seg2->partner;
		}
	}
	else if (!(s21 + s22))
	{
		if (s22 < 0)
		{
			seg1 = seg1->partner;
		}
		if (s11 < 0 || s12 < 0)
		{
			seg2 = seg2->partner;
		}
	}
	else
	{
		if (s21 < 0 || s22 < 0)
		{
			seg1 = seg1->partner;
		}
		if (s11 < 0 || s12 < 0)
		{
			seg2 = seg2->partner;
		}
	}

	if (seg1->subsector == seg2->subsector)
	{
		return true;
	}

	checks1 = seg1;
	checks2 = seg2;

#ifdef TEST_SPLIT
	TVec s1dir = Normalize(checkv12 - checkv11);
	TVec s2dir = Normalize(checkv22 - checkv21);

	float s1len = Length(checkv12 - checkv11);
	checkv11 = *seg1->v1;
	do
	{
		if (s1len > split_factor)
		{
			checkv12 = checkv11 + split_factor * s1dir;
double dist = DotProduct(checkv12, seg1->normal) - seg1->dist;
checkv12 -= dist * seg1->normal;
		}
		else
		{
			checkv12 = *seg1->v2;
		}

		float s2len = Length(checkv22 - checkv21);
		checkv21 = *seg2->v1;
		do
		{
			if (s2len > split_factor)
			{
				checkv22 = checkv21 + split_factor * s2dir;
double dist = DotProduct(checkv22, seg2->normal) - seg2->dist;
checkv22 -= dist * seg2->normal;
			}
			else
			{
				checkv22 = *seg2->v2;
			}

			if (CheckSegParts())
			{
				return true;
			}

			checkv21 = checkv22;
			s2len -= split_factor;

		} while (s2len > split_factor);

		checkv11 = checkv12;
		s1len -= split_factor;

	} while (s1len > split_factor);

	return false;
#else
	checkv11 = *seg1->v1;
	checkv12 = *seg1->v2;
	checkv21 = *seg2->v1;
	checkv22 = *seg2->v2;

	return CheckSegParts();
#endif
}

//==========================================================================
//
//	SetAccept
//
//==========================================================================

static void	SetAccept(subsector_t *ss1, subsector_t *ss2)
{
	int s1 = ss1 - subsectors;
	int s2 = ss2 - subsectors;
	set_vis(s1, s2);
	set_vis(s2, s1);
}

//==========================================================================
//
//	GetVis
//
//==========================================================================

static byte GetVis(subsector_t *ss1, subsector_t *ss2)
{
	int s1 = ss1 - subsectors;
	int s2 = ss2 - subsectors;
	return get_vis(s1, s2);
}

//==========================================================================
//
//	MarkSegs
//
//==========================================================================

static void MarkSegs(seg_t *seg1, seg_t *seg2)
{
	if (seg1 == seg2->partner || seg2 == seg1->partner)
	{
		//	Already checked
		return;
	}

	if (seg1->linedef && seg1->linedef == seg2->linedef)
	{
		//	Segs belongs to the same line, no los possible
		return;
	}

	if (GetVis(seg1->subsector, seg2->subsector) &&
		GetVis(seg1->partner->subsector, seg2->subsector) &&
		GetVis(seg1->subsector, seg2->partner->subsector) &&
		GetVis(seg1->partner->subsector, seg2->partner->subsector))
	{
		//	All subsector pairs already are known as visible
		return;
	}

	if (CheckSegs(seg1, seg2))
	{
		//	Los possible, mark all subsector pairs
		SetAccept(seg1->subsector, seg2->subsector);
		SetAccept(seg1->partner->subsector, seg2->subsector);
		SetAccept(seg1->subsector, seg2->partner->subsector);
		SetAccept(seg1->partner->subsector, seg2->partner->subsector);
	}
}

//==========================================================================
//
//	BuildPVS
//
//==========================================================================

void BuildPVS(void)
{
	int i, j;

	if (!silent_mode)
	{
		cerr << "Creating vis data ... ";
	}

	validcount = 0;
	vis = new byte[numsubsectors * numsubsectors];

	for (i = 0; i < numsubsectors; i++)
	{
		set_vis(i, i);
	}

	for (i = 0; i < numsegs; i++)
	{
		seg_t *seg = &segs[i];
		if (!seg->subsector)
		{
			Error("Seg without subsector");
		}
		if (seg->partner)
		{
			SetAccept(seg->subsector, seg->partner->subsector);
			SetAccept(seg->partner->subsector, seg->subsector);
		}
	}

	//	Check all seg pairs
	for (i = 0; i < numsegs; i++)
	{
		if (!silent_mode && (!i || (numsegs - i) % 10 == 0))
		{
			fprintf(stderr, "%04d\b\b\b\b", numsegs - i);
		}
		seg_t *seg1 = &segs[i];

		if (seg1->partner)
		{
			if (seg1->partner < seg1)
			{
				//	Already checked partner seg
				continue;
			}
			for (j = i + 1; j < numsegs; j++)
			{
				seg_t *seg2 = &segs[j];

				if (seg2->partner)
				{
					if (seg2->partner < seg2)
					{
						//	Already checked	partner seg
						continue;
					}
					MarkSegs(seg1, seg2);
				}
			}
		}
	}

	//	Create packed data
	int acount = 0;
	int rcount = 0;
	int	pak_w = (numsubsectors + 7) >> 3;
	byte *packed_vis = new byte[pak_w * numsubsectors];
	for (i = 0; i < numsubsectors; i++)
	{
		byte *currvis = packed_vis + i * pak_w;
		for (j = 0; j < numsubsectors; j++)
		{
			if (get_vis(i, j))
			{
				currvis[j >> 3] |= 1 << (j & 7);
				acount++;
			}
			else
			{
				rcount++;
			}
		}
	}

	if (!silent_mode)
	{
		cerr << acount << " accepts, " << rcount << " rejects\n";
	}

	//	Write lump
	outwad.AddLump("GL_PVS", packed_vis, pak_w * numsubsectors);

	delete packed_vis;
	delete vis;
}

